@@ -223,9 +223,24 @@ class _SourceMessageReceiver:
     def __init__(self, edge: Edge, node: Node) -> None:
         self.__delete_message_queue = _DeleteMessageQueue(edge, node)
 
+        async def handle_received_message(
+            message: Message, receipt_handle: str
+        ) -> bool:
+            try:
+                await node.handle_received_message(message=message, source=edge.name)
+            except asyncio.CancelledError:
+                raise
+            except Exception:
+                getLogger().exception(
+                    f"Error handling recevied message for {edge.name}"
+                )
+                return False
+            else:
+                self.__delete_message_queue.put_nowait(receipt_handle)
+            return True
+
         async def receive() -> None:
             getLogger().info(f"Receiving messages from {edge.name}")
-            error_count = 0
             while True:
                 try:
                     response = await asyncio.get_running_loop().run_in_executor(
@@ -239,53 +254,45 @@ class _SourceMessageReceiver:
                             WaitTimeSeconds=20,
                         ),
                     )
-                    error_count = 0
                 except asyncio.CancelledError:
                     raise
                 except catch_aws_error("AWS.SimpleQueueService.NonExistentQueue"):
                     getLogger().warning(f"Queue {edge.queue} does not exist, exiting")
                     break
                 except Exception:
-                    error_count += 1
-                    if error_count == 10:
-                        getLogger().critical(
-                            f"Recevied 10 errors in a row trying to receive from {edge.queue}, exiting"
-                        )
-                        raise
-                    else:
-                        getLogger().exception(
-                            f"Error receiving messages from {edge.name}, retrying"
-                        )
-                        await asyncio.sleep(10)
+                    getLogger().exception(
+                        f"Error receiving messages from {edge.name}, retrying"
+                    )
+                    await asyncio.sleep(20)
                 else:
                     if not (sqs_messages := response.get("Messages")):
                         continue
                     getLogger().info(f"Received {len(sqs_messages)} from {edge.name}")
-                    for sqs_message in sqs_messages:
-                        message = Message(
-                            body=sqs_message["Body"],
-                            group_id=sqs_message["Attributes"]["MessageGroupId"],
-                            message_type=node.receive_message_type,
-                            tracking_id=sqs_message["MessageAttributes"]
-                            .get("trackingId", {})
-                            .get("StringValue"),
-                            previous_tracking_ids=sqs_message["MessageAttributes"]
-                            .get("prevTrackingIds", {})
-                            .get("StringValue"),
+
+                    handle_received_messages = [
+                        handle_received_message(
+                            Message(
+                                body=sqs_message["Body"],
+                                group_id=sqs_message["Attributes"]["MessageGroupId"],
+                                message_type=node.receive_message_type,
+                                tracking_id=sqs_message["MessageAttributes"]
+                                .get("trackingId", {})
+                                .get("StringValue"),
+                                previous_tracking_ids=sqs_message["MessageAttributes"]
+                                .get("prevTrackingIds", {})
+                                .get("StringValue"),
+                            ),
+                            sqs_message["ReceiptHandle"],
                         )
-                        receipt_handle = sqs_message["ReceiptHandle"]
-                        try:
-                            await node.handle_received_message(
-                                message=message, source=edge.name
-                            )
-                        except asyncio.CancelledError:
-                            raise
-                        except Exception:
-                            getLogger().exception(
-                                f"Error handling recevied message for {edge.name}"
-                            )
-                        else:
-                            self.__delete_message_queue.put_nowait(receipt_handle)
+                        for sqs_message in sqs_messages
+                    ]
+                    if node._concurrent_processing:
+                        await asyncio.gather(*handle_received_messages)
+                    else:
+                        for coro in handle_received_messages:
+                            if not await coro:
+                                break
+
             getLogger().info(f"Stopping receiving messages from {edge.name}")
 
         self.__task = asyncio.create_task(
@@ -390,11 +397,19 @@ class CognitoAIOHTTPTransport(AIOHTTPTransport):
 
 
 class Node(BaseNode):
+    """
+    Base class for all implemented asyncio Nodes.
+
+    Nodes of this class must be instantiated outside of
+    the asyncio event loop.
+    """
+
     def __init__(
         self,
         *,
         appsync_endpoint: str = None,
         client_id: str = None,
+        concurrent_processing: bool = False,
         name: str = None,
         password: str = None,
         tenant: str = None,
@@ -416,9 +431,14 @@ class Node(BaseNode):
         self.__audit_records_queues: dict[str, _AuditRecordQueue] = dict()
         self.__bulk_data_storage_queue: _BulkDataStorageQueue = None
         self.__lock: asyncio.Lock = None
+        self.__concurrent_processing = concurrent_processing
         self.__source_message_receivers: list[_SourceMessageReceiver] = list()
         self.__target_message_queues: dict[str, _TargetMessageQueue] = dict()
 
+    @property
+    def _concurrent_processing(self) -> bool:
+        return self.__concurrent_processing
+
     @property
     def _lock(self) -> asyncio.Lock:
         return self.__lock
@@ -431,6 +451,12 @@ class Node(BaseNode):
         extra_attributes: dict[str, Any] = None,
         source: str = None,
     ) -> None:
+        """
+        Audits the provided message. If extra_attibutes is
+        supplied, they will be added to the message's audit
+        dict. If source is provided, it will be recorded in
+        the audit.
+        """
         extra_attributes = extra_attributes or dict()
         message_type = message.message_type
         record = dict(
@@ -449,12 +475,25 @@ class Node(BaseNode):
             raise ValueError(f"Unrecognized message type {message_type.name}")
 
     async def handle_bulk_data(self, data: Union[bytearray, bytes]) -> str:
+        """
+        Posts data as bulk data and returns a GET URL for data retrieval.
+        Normally this returned URL will be used as a "ticket" in messages
+        that require bulk data.
+        """
         return await (await self.__bulk_data_storage_queue.get()).handle_bulk_data(data)
 
     async def handle_received_message(self, *, message: Message, source: str) -> None:
+        """
+        Callback called when a message is received. Subclasses that receive messages
+        should override this method.
+        """
         pass
 
     async def join(self) -> None:
+        """
+        Joins the calling thread with this Node. Will block until all
+        join conditions are satified.
+        """
         await asyncio.gather(
             *[
                 source_message_receiver.join()
@@ -471,11 +510,19 @@ class Node(BaseNode):
             await audit_records_queue.join()
 
     def send_message(self, /, message: Message, *, targets: set[Edge] = None) -> None:
+        """
+        Send the message to the specified targets. If no targets are specified
+        the message will be sent to all targets.
+        """
         self.send_messages([message], targets=targets)
 
     def send_messages(
         self, /, messages: list[Message], *, targets: set[Edge] = None
     ) -> None:
+        """
+        Send the messages to the specified targets. If no targets are specified
+        the messages will be sent to all targets.
+        """
         if messages:
             for target in targets or self.targets:
                 if target_message_queue := self.__target_message_queues.get(
@@ -487,6 +534,9 @@ class Node(BaseNode):
                     getLogger().warning(f"Target {target.name} does not exist")
 
     async def start(self) -> None:
+        """
+        Starts this Node. Must be called prior to any other usage.
+        """
         getLogger().info(f"Starting Node {self.name}")
         self.__lock = asyncio.Lock()
         self.__bulk_data_storage_queue = _BulkDataStorageQueue(self)
@@ -541,5 +591,6 @@ class Node(BaseNode):
         ]
 
     async def start_and_run_forever(self) -> None:
+        """Will start this Node and run until the containing Task is cancelled"""
         await self.start()
         await self.join()
