@@ -1,7 +1,9 @@
 from __future__ import annotations
 
 import json
+from concurrent.futures import Executor, ThreadPoolExecutor, as_completed, wait
 from datetime import datetime, timezone
+from functools import partial
 from gzip import GzipFile
 from io import BytesIO
 from queue import Empty, Queue
@@ -184,6 +186,10 @@ class _TargetMessageQueue(Queue):
 
 
 class Node(BaseNode):
+    """
+    Base class for all threading Nodes.
+    """
+
     def __init__(
         self,
         *,
@@ -223,6 +229,12 @@ class Node(BaseNode):
         extra_attributes: dict[str, Any] = None,
         source: str = None,
     ) -> None:
+        """
+        Audits the provided message. If extra_attibutes is
+        supplied, they will be added to the message's audit
+        dict. If source is provided, it will be recorded in
+        the audit.
+        """
         extra_attributes = extra_attributes or dict()
         message_type = message.message_type
         record = dict(
@@ -241,23 +253,44 @@ class Node(BaseNode):
             raise ValueError(f"Unrecognized message type {message_type.name}")
 
     def handle_bulk_data(self, data: Union[bytearray, bytes]) -> str:
+        """
+        Posts data as bulk data and returns a GET URL for data retrieval.
+        Normally this returned URL will be used as a "ticket" in messages
+        that require bulk data.
+        """
         return self.__bulk_data_storage_queue.get().handle_bulk_data(data)
 
     def handle_received_message(self, *, message: Message, source: str) -> None:
+        """
+        Callback called when a message is received. Subclasses that receive messages
+        should override this method.
+        """
         pass
 
     def join(self) -> None:
+        """
+        Joins the calling thread with this Node. Will block until all
+        join conditions are satified.
+        """
         for target_message_queue in self.__target_message_queues.values():
             target_message_queue.join()
         for audit_records_queue in self.__audit_records_queues.values():
             audit_records_queue.join()
 
     def send_message(self, /, message: Message, *, targets: set[Edge] = None) -> None:
+        """
+        Send the message to the specified targets. If no targets are specified
+        the message will be sent to all targets.
+        """
         self.send_messages([message], targets=targets)
 
     def send_messages(
         self, /, messages: list[Message], *, targets: set[Edge] = None
     ) -> None:
+        """
+        Send the messages to the specified targets. If no targets are specified
+        the messages will be sent to all targets.
+        """
         if messages:
             for target in targets or self.targets:
                 if target_message_queue := self.__target_message_queues.get(
@@ -269,6 +302,9 @@ class Node(BaseNode):
                     getLogger().warning(f"Target {target.name} does not exist")
 
     def start(self) -> None:
+        """
+        Starts this Node. Must be called prior to any other usage.
+        """
         getLogger().info(f"Starting Node {self.name}")
         with self._lock:
             with self._gql_client as session:
@@ -316,6 +352,7 @@ class Node(BaseNode):
         }
 
     def stop(self) -> None:
+        """Stops the Node's processing."""
         pass
 
 
@@ -363,15 +400,26 @@ class _DeleteMessageQueue(Queue):
 
 
 class _SourceMessageReceiver(Thread):
-    def __init__(self, edge: Edge, node: Node) -> None:
+    def __init__(self, edge: Edge, node: AppNode) -> None:
         self.__continue = Event()
         self.__continue.set()
         self.__delete_message_queue = _DeleteMessageQueue(edge, node)
 
+        def handle_received_message(message: Message, receipt_handle: str) -> bool:
+            try:
+                node.handle_received_message(message=message, source=edge.name)
+            except Exception:
+                getLogger().exception(
+                    f"Error handling recevied message for {edge.name}"
+                )
+                return False
+            else:
+                self.__delete_message_queue.put_nowait(receipt_handle)
+            return True
+
         def receive() -> None:
             self.__continue.wait()
             getLogger().info(f"Receiving messages from {edge.name}")
-            error_count = 0
             while self.__continue.is_set():
                 try:
                     response = node._sqs_client.receive_message(
@@ -381,49 +429,47 @@ class _SourceMessageReceiver(Thread):
                         QueueUrl=edge.queue,
                         WaitTimeSeconds=20,
                     )
-                    error_count = 0
                 except catch_aws_error("AWS.SimpleQueueService.NonExistentQueue"):
                     getLogger().warning(f"Queue {edge.queue} does not exist, exiting")
                     break
                 except Exception:
-                    error_count += 1
-                    if error_count == 10:
-                        getLogger().critical(
-                            f"Recevied 10 errors in a row trying to receive from {edge.queue}, exiting"
-                        )
-                        raise
-                    else:
-                        getLogger().exception(
-                            f"Error receiving messages from {edge.name}, retrying"
-                        )
-                        sleep(10)
+                    getLogger().exception(
+                        f"Error receiving messages from {edge.name}, retrying"
+                    )
+                    sleep(20)
                 else:
                     if not (sqs_messages := response.get("Messages")):
                         continue
                     getLogger().info(f"Received {len(sqs_messages)} from {edge.name}")
-                    for sqs_message in sqs_messages:
-                        message = Message(
-                            body=sqs_message["Body"],
-                            group_id=sqs_message["Attributes"]["MessageGroupId"],
-                            message_type=node.receive_message_type,
-                            tracking_id=sqs_message["MessageAttributes"]
-                            .get("trackingId", {})
-                            .get("StringValue"),
-                            previous_tracking_ids=sqs_message["MessageAttributes"]
-                            .get("prevTrackingIds", {})
-                            .get("StringValue"),
+
+                    handle_received_messages = [
+                        partial(
+                            handle_received_message,
+                            Message(
+                                body=sqs_message["Body"],
+                                group_id=sqs_message["Attributes"]["MessageGroupId"],
+                                message_type=node.receive_message_type,
+                                tracking_id=sqs_message["MessageAttributes"]
+                                .get("trackingId", {})
+                                .get("StringValue"),
+                                previous_tracking_ids=sqs_message["MessageAttributes"]
+                                .get("prevTrackingIds", {})
+                                .get("StringValue"),
+                            ),
+                            sqs_message["ReceiptHandle"],
                         )
-                        receipt_handle = sqs_message["ReceiptHandle"]
-                        try:
-                            node.handle_received_message(
-                                message=message, source=edge.name
-                            )
-                        except Exception:
-                            getLogger().exception(
-                                f"Error handling recevied message for {edge.name}"
-                            )
-                        else:
-                            self.__delete_message_queue.put_nowait(receipt_handle)
+                        for sqs_message in sqs_messages
+                    ]
+
+                    if executor := node._executor:
+                        wait(
+                            [executor.submit(func) for func in handle_received_messages]
+                        )
+                    else:
+                        for func in handle_received_messages:
+                            if not func():
+                                break
+
             getLogger().info(f"Stopping receiving messages from {edge.name}")
 
         super().__init__(name=f"SourceMessageReceiver({edge.name})", target=receive)
@@ -438,11 +484,17 @@ class _SourceMessageReceiver(Thread):
 
 
 class AppNode(Node):
+    """
+    A daemon Node intended to be used as either a stand-alone application
+    or as a part of a larger application.
+    """
+
     def __init__(
         self,
         *,
         appsync_endpoint: str = None,
         client_id: str = None,
+        executor: Executor = None,
         name: str = None,
         password: str = None,
         tenant: str = None,
@@ -460,16 +512,27 @@ class AppNode(Node):
             user_pool_id=user_pool_id,
             username=username,
         )
+        self.__executor = executor
         self.__source_message_receivers: list[_SourceMessageReceiver] = list()
         self.__stop = Event()
 
+    @property
+    def _executor(self) -> Executor:
+        return self.__executor
+
     def join(self) -> None:
+        """
+        Method to join all the app node receivers so that main thread can wait for their execution to complete.
+        """
         self.__stop.wait()
         for app_node_receiver in self.__source_message_receivers:
             app_node_receiver.join()
         super().join()
 
     def start(self) -> None:
+        """
+        Calls start of Node class
+        """
         super().start()
         self.__stop.clear()
         self.__source_message_receivers = [
@@ -477,23 +540,35 @@ class AppNode(Node):
         ]
 
     def start_and_run_forever(self) -> None:
+        """Will start this Node and run until stop is called"""
         self.start()
         self.join()
 
     def stop(self) -> None:
+        """
+        Stops the Node gracefully
+        """
         self.__stop.set()
         for app_node_receiver in self.__source_message_receivers:
             app_node_receiver.stop()
 
 
 class LambdaNode(Node):
+    """
+    A Node class intended to be implemented in an AWS Lambda function.
+    Nodes that inherit from this class are automatically started on
+    creation.
+    """
+
     def __init__(
         self,
         *,
         appsync_endpoint: str = None,
         client_id: str = None,
+        concurrent_processing: bool = False,
         name: str = None,
         password: str = None,
+        report_batch_item_failures: bool = False,
         tenant: str = None,
         timeout: float = None,
         user_pool_id: str = None,
@@ -510,14 +585,25 @@ class LambdaNode(Node):
             username=username,
         )
         self.start()
+        self.__executor: Executor = (
+            ThreadPoolExecutor() if concurrent_processing else None
+        )
         self.__queue_name_to_source = {
             edge.queue.split("/")[-1:][0]: edge.name for edge in self._sources
         }
+        self.__report_batch_item_failures = report_batch_item_failures
 
     def _get_source(self, queue_arn: str) -> str:
         return self.__queue_name_to_source[queue_arn.split(":")[-1:][0]]
 
     def handle_event(self, event: LambdaEvent) -> None:
+        """
+        Handles the AWS Lambda event passed into the containing
+        AWS Lambda function during invocation.
+
+        This is intended to be the only called method in your
+        containing AWS Lambda function.
+        """
         records: list[
             dict[
                 str,
@@ -534,13 +620,30 @@ class LambdaNode(Node):
         if not (records := event.get("Records")):
             getLogger().warning(f"No Records found in event {event}")
             return
-        source: str = None
-        try:
-            for record in records:
-                if not source:
-                    source = self._get_source(record["eventSourceARN"])
-                    getLogger().info(f"Received {len(records)} messages from {source}")
-                message = Message(
+
+        source = self._get_source(records[0]["eventSourceARN"])
+        getLogger().info(f"Received {len(records)} messages from {source}")
+        batch_item_failures: list[str] = (
+            [record["messageId"] for record in records]
+            if self.__report_batch_item_failures
+            else None
+        )
+
+        def handle_received_message(message: Message, message_id: str) -> None:
+            try:
+                self.handle_received_message(message=message, source=source)
+            except Exception:
+                if not self.__report_batch_item_failures:
+                    raise
+                getLogger().exception(f"Error handling recevied message for {source}")
+            else:
+                if self.__report_batch_item_failures:
+                    batch_item_failures.remove(message_id)
+
+        handle_received_messages = [
+            partial(
+                handle_received_message,
+                Message(
                     body=record["body"],
                     group_id=record["attributes"]["MessageGroupId"],
                     message_type=self.receive_message_type,
@@ -551,7 +654,28 @@ class LambdaNode(Node):
                     .get("trackingId", {})
                     .get("stringValue")
                     or uuid4().hex,
-                )
-                self.handle_received_message(message=message, source=source)
+                ),
+                record["messageId"],
+            )
+            for record in records
+        ]
+
+        try:
+            if executor := self.__executor:
+                for future in as_completed(
+                    [executor.submit(func) for func in handle_received_messages]
+                ):
+                    if exception := future.exception():
+                        raise exception
+            else:
+                for func in handle_received_messages:
+                    func()
         finally:
             self.join()
+        if self.__report_batch_item_failures and batch_item_failures:
+            return dict(
+                batchItemFailures=[
+                    dict(itemIdentifier=message_id)
+                    for message_id in batch_item_failures
+                ]
+            )
